@@ -10,6 +10,7 @@
  */
 
 #include "mozilla/Attributes.h"
+#include "mozilla/CheckedInt.h"
 #include "mozilla/MathAlgorithms.h"
 #include "mozilla/Mutex.h"
 #include "mozilla/SystemGroup.h"
@@ -17,6 +18,7 @@
 #include "base/basictypes.h"
 
 #include "nsMultiplexInputStream.h"
+#include "nsIBufferedStreams.h"
 #include "nsICloneableInputStream.h"
 #include "nsIMultiplexInputStream.h"
 #include "nsISeekableStream.h"
@@ -26,6 +28,9 @@
 #include "nsIIPCSerializableInputStream.h"
 #include "mozilla/ipc/InputStreamUtils.h"
 #include "nsIAsyncInputStream.h"
+#include "nsIInputStreamLength.h"
+#include "nsNetUtil.h"
+#include "nsStreamUtils.h"
 
 using namespace mozilla;
 using namespace mozilla::ipc;
@@ -40,7 +45,9 @@ class nsMultiplexInputStream final : public nsIMultiplexInputStream,
                                      public nsIIPCSerializableInputStream,
                                      public nsICloneableInputStream,
                                      public nsIAsyncInputStream,
-                                     public nsIInputStreamCallback {
+                                     public nsIInputStreamCallback,
+                                     public nsIInputStreamLength,
+                                     public nsIAsyncInputStreamLength {
  public:
   nsMultiplexInputStream();
 
@@ -48,18 +55,27 @@ class nsMultiplexInputStream final : public nsIMultiplexInputStream,
   NS_DECL_NSIINPUTSTREAM
   NS_DECL_NSIMULTIPLEXINPUTSTREAM
   NS_DECL_NSISEEKABLESTREAM
+  NS_DECL_NSITELLABLESTREAM
   NS_DECL_NSIIPCSERIALIZABLEINPUTSTREAM
   NS_DECL_NSICLONEABLEINPUTSTREAM
   NS_DECL_NSIASYNCINPUTSTREAM
   NS_DECL_NSIINPUTSTREAMCALLBACK
+  NS_DECL_NSIINPUTSTREAMLENGTH
+  NS_DECL_NSIASYNCINPUTSTREAMLENGTH
 
+  // This is used for nsIAsyncInputStream::AsyncWait
   void AsyncWaitCompleted();
 
+  // This is used for nsIAsyncInputStreamLength::AsyncLengthWait
+  void AsyncWaitCompleted(int64_t aLength, const MutexAutoLock& aProofOfLock);
+
   struct StreamData {
-    void Initialize(nsIInputStream* aStream) {
+    void Initialize(nsIInputStream* aStream, bool aBuffered) {
       mStream = aStream;
       mAsyncStream = do_QueryInterface(aStream);
       mSeekableStream = do_QueryInterface(aStream);
+      mTellableStream = do_QueryInterface(aStream);
+      mBuffered = aBuffered;
     }
 
     nsCOMPtr<nsIInputStream> mStream;
@@ -68,15 +84,22 @@ class nsMultiplexInputStream final : public nsIMultiplexInputStream,
     nsCOMPtr<nsIAsyncInputStream> mAsyncStream;
     // This can be null.
     nsCOMPtr<nsISeekableStream> mSeekableStream;
+    // This can be null.
+    nsCOMPtr<nsITellableStream> mTellableStream;
+
+    // True if the stream is wrapped with nsIBufferedInputStream.
+    bool mBuffered;
   };
 
+  Mutex& GetLock() { return mLock; }
+
  private:
   ~nsMultiplexInputStream() {}
 
   nsresult AsyncWaitInternal();
 
-  // This method updates mSeekableStreams, mIPCSerializableStreams,
-  // mCloneableStreams and mAsyncInputStreams values.
+  // This method updates mSeekableStreams, mTellableStreams,
+  // mIPCSerializableStreams, mCloneableStreams and mAsyncInputStreams values.
   void UpdateQIMap(StreamData& aStream, int32_t aCount);
 
   struct MOZ_STACK_CLASS ReadSegmentsState {
@@ -87,14 +110,23 @@ class nsMultiplexInputStream final : public nsIMultiplexInputStream,
     bool mDone;
   };
 
+  template <typename M>
+  void SerializeInternal(InputStreamParams& aParams,
+                         FileDescriptorArray& aFileDescriptors,
+                         bool aDelayedStart, uint32_t aMaxSize,
+                         uint32_t* aSizeUsed, M* aManager);
+
   static nsresult ReadSegCb(nsIInputStream* aIn, void* aClosure,
                             const char* aFromRawSegment, uint32_t aToOffset,
                             uint32_t aCount, uint32_t* aWriteCount);
 
   bool IsSeekable() const;
+  bool IsTellable() const;
   bool IsIPCSerializable() const;
   bool IsCloneable() const;
   bool IsAsyncInputStream() const;
+  bool IsInputStreamLength() const;
+  bool IsAsyncInputStreamLength() const;
 
   Mutex mLock;  // Protects access to all data members.
 
@@ -107,11 +139,18 @@ class nsMultiplexInputStream final : public nsIMultiplexInputStream,
   uint32_t mAsyncWaitFlags;
   uint32_t mAsyncWaitRequestedCount;
   nsCOMPtr<nsIEventTarget> mAsyncWaitEventTarget;
+  nsCOMPtr<nsIInputStreamLengthCallback> mAsyncWaitLengthCallback;
+
+  class AsyncWaitLengthHelper;
+  RefPtr<AsyncWaitLengthHelper> mAsyncWaitLengthHelper;
 
   uint32_t mSeekableStreams;
+  uint32_t mTellableStreams;
   uint32_t mIPCSerializableStreams;
   uint32_t mCloneableStreams;
   uint32_t mAsyncInputStreams;
+  uint32_t mInputStreamLengths;
+  uint32_t mAsyncInputStreamLengths;
 };
 
 NS_IMPL_ADDREF(nsMultiplexInputStream)
@@ -124,18 +163,24 @@ NS_INTERFACE_MAP_BEGIN(nsMultiplexInputStream)
   NS_INTERFACE_MAP_ENTRY(nsIMultiplexInputStream)
   NS_INTERFACE_MAP_ENTRY(nsIInputStream)
   NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsISeekableStream, IsSeekable())
+  NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsITellableStream, IsTellable())
   NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsIIPCSerializableInputStream,
                                      IsIPCSerializable())
   NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsICloneableInputStream, IsCloneable())
   NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsIAsyncInputStream, IsAsyncInputStream())
   NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsIInputStreamCallback,
                                      IsAsyncInputStream())
+  NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsIInputStreamLength,
+                                     IsInputStreamLength())
+  NS_INTERFACE_MAP_ENTRY_CONDITIONAL(nsIAsyncInputStreamLength,
+                                     IsAsyncInputStreamLength())
   NS_INTERFACE_MAP_ENTRY_AMBIGUOUS(nsISupports, nsIMultiplexInputStream)
   NS_IMPL_QUERY_CLASSINFO(nsMultiplexInputStream)
 NS_INTERFACE_MAP_END
 
 NS_IMPL_CI_INTERFACE_GETTER(nsMultiplexInputStream, nsIMultiplexInputStream,
-                            nsIInputStream, nsISeekableStream)
+                            nsIInputStream, nsISeekableStream,
+                            nsITellableStream)
 
 static nsresult AvailableMaybeSeek(nsMultiplexInputStream::StreamData& aStream,
                                    uint64_t* aResult) {
@@ -156,16 +201,17 @@ static nsresult AvailableMaybeSeek(nsMultiplexInputStream::StreamData& aStream,
   return rv;
 }
 
-static nsresult TellMaybeSeek(nsISeekableStream* aSeekable, int64_t* aResult) {
-  nsresult rv = aSeekable->Tell(aResult);
-  if (rv == NS_BASE_STREAM_CLOSED) {
+static nsresult TellMaybeSeek(nsITellableStream* aTellable,
+                              nsISeekableStream* aSeekable, int64_t* aResult) {
+  nsresult rv = aTellable->Tell(aResult);
+  if (rv == NS_BASE_STREAM_CLOSED && aSeekable) {
     // Blindly seek to the current position if Tell() returns
     // NS_BASE_STREAM_CLOSED.
     // If nsIFileInputStream is closed in Read() due to CLOSE_ON_EOF flag,
     // Seek() could reopen the file if REOPEN_ON_REWIND flag is set.
     nsresult rvSeek = aSeekable->Seek(nsISeekableStream::NS_SEEK_CUR, 0);
     if (NS_SUCCEEDED(rvSeek)) {
-      rv = aSeekable->Tell(aResult);
+      rv = aTellable->Tell(aResult);
     }
   }
   return rv;
@@ -179,9 +225,12 @@ nsMultiplexInputStream::nsMultiplexInputStream()
       mAsyncWaitFlags(0),
       mAsyncWaitRequestedCount(0),
       mSeekableStreams(0),
+      mTellableStreams(0),
       mIPCSerializableStreams(0),
       mCloneableStreams(0),
-      mAsyncInputStreams(0) {}
+      mAsyncInputStreams(0),
+      mInputStreamLengths(0),
+      mAsyncInputStreamLengths(0) {}
 
 NS_IMETHODIMP
 nsMultiplexInputStream::GetCount(uint32_t* aCount) {
@@ -192,14 +241,26 @@ nsMultiplexInputStream::GetCount(uint32_t* aCount) {
 
 NS_IMETHODIMP
 nsMultiplexInputStream::AppendStream(nsIInputStream* aStream) {
+  nsCOMPtr<nsIInputStream> stream = aStream;
+
+  bool buffered = false;
+  if (!NS_InputStreamIsBuffered(stream)) {
+    nsCOMPtr<nsIInputStream> bufferedStream;
+    nsresult rv = NS_NewBufferedInputStream(getter_AddRefs(bufferedStream),
+                                            stream.forget(), 4096);
+    NS_ENSURE_SUCCESS(rv, rv);
+    stream = bufferedStream.forget();
+    buffered = true;
+  }
+
   MutexAutoLock lock(mLock);
 
-  StreamData* streamData = mStreams.AppendElement();
+  StreamData* streamData = mStreams.AppendElement(fallible);
   if (NS_WARN_IF(!streamData)) {
     return NS_ERROR_OUT_OF_MEMORY;
   }
 
-  streamData->Initialize(aStream);
+  streamData->Initialize(stream, buffered);
 
   UpdateQIMap(*streamData, 1);
 
@@ -222,6 +283,17 @@ nsMultiplexInputStream::GetStream(uint32_t aIndex, nsIInputStream** aResult) {
   StreamData& streamData = mStreams.ElementAt(aIndex);
 
   nsCOMPtr<nsIInputStream> stream = streamData.mStream;
+
+  if (streamData.mBuffered) {
+    nsCOMPtr<nsIBufferedInputStream> bufferedStream = do_QueryInterface(stream);
+    MOZ_ASSERT(bufferedStream);
+
+    nsresult rv = bufferedStream->GetData(getter_AddRefs(stream));
+    if (NS_WARN_IF(NS_FAILED(rv))) {
+      return rv;
+    }
+  }
+
   stream.forget(aResult);
   return NS_OK;
 }
@@ -237,7 +309,10 @@ nsMultiplexInputStream::Close() {
     MutexAutoLock lock(mLock);
     uint32_t len = mStreams.Length();
     for (uint32_t i = 0; i < len; ++i) {
-      streams.AppendElement(mStreams[i].mStream);
+      if (NS_WARN_IF(!streams.AppendElement(mStreams[i].mStream, fallible))) {
+        mStatus = NS_BASE_STREAM_CLOSED;
+        return NS_ERROR_OUT_OF_MEMORY;
+      }
     }
     mStatus = NS_BASE_STREAM_CLOSED;
   }
@@ -342,8 +417,9 @@ nsMultiplexInputStream::Read(char* aBuf, uint32_t aCount, uint32_t* aResult) {
     // XXX some streams return NS_BASE_STREAM_CLOSED to indicate EOF.
     // (This is a bug in those stream implementations)
     if (rv == NS_BASE_STREAM_CLOSED) {
-      NS_NOTREACHED(
-          "Input stream's Read method returned NS_BASE_STREAM_CLOSED");
+      MOZ_ASSERT_UNREACHABLE(
+          "Input stream's Read method returned "
+          "NS_BASE_STREAM_CLOSED");
       rv = NS_OK;
       read = 0;
     } else if (NS_FAILED(rv)) {
@@ -396,8 +472,9 @@ nsMultiplexInputStream::ReadSegments(nsWriteSegmentFun aWriter, void* aClosure,
     // XXX some streams return NS_BASE_STREAM_CLOSED to indicate EOF.
     // (This is a bug in those stream implementations)
     if (rv == NS_BASE_STREAM_CLOSED) {
-      NS_NOTREACHED(
-          "Input stream's Read method returned NS_BASE_STREAM_CLOSED");
+      MOZ_ASSERT_UNREACHABLE(
+          "Input stream's Read method returned "
+          "NS_BASE_STREAM_CLOSED");
       rv = NS_OK;
       read = 0;
     }
@@ -485,6 +562,8 @@ nsMultiplexInputStream::Seek(int32_t aWhence, int64_t aOffset) {
         return NS_ERROR_FAILURE;
       }
 
+      MOZ_ASSERT(mStreams[i].mTellableStream);
+
       // See if all remaining streams should be rewound
       if (remaining == 0) {
         if (i < oldCurrentStream ||
@@ -505,7 +584,7 @@ nsMultiplexInputStream::Seek(int32_t aWhence, int64_t aOffset) {
           (i == oldCurrentStream && !oldStartedReadingCurrent)) {
         streamPos = 0;
       } else {
-        rv = TellMaybeSeek(stream, &streamPos);
+        rv = TellMaybeSeek(mStreams[i].mTellableStream, stream, &streamPos);
         if (NS_WARN_IF(NS_FAILED(rv))) {
           return rv;
         }
@@ -588,7 +667,8 @@ nsMultiplexInputStream::Seek(int32_t aWhence, int64_t aOffset) {
     int64_t remaining = -aOffset;
     for (uint32_t i = mCurrentStream; remaining && i != (uint32_t)-1; --i) {
       int64_t pos;
-      rv = TellMaybeSeek(mStreams[i].mSeekableStream, &pos);
+      rv = TellMaybeSeek(mStreams[i].mTellableStream,
+                         mStreams[i].mSeekableStream, &pos);
       if (NS_WARN_IF(NS_FAILED(rv))) {
         return rv;
       }
@@ -667,7 +747,7 @@ nsMultiplexInputStream::Seek(int32_t aWhence, int64_t aOffset) {
           remaining += streamPos;
         } else {
           int64_t avail;
-          rv = TellMaybeSeek(stream, &avail);
+          rv = TellMaybeSeek(mStreams[i].mTellableStream, stream, &avail);
           if (NS_WARN_IF(NS_FAILED(rv))) {
             return rv;
           }
@@ -711,12 +791,13 @@ nsMultiplexInputStream::Tell(int64_t* aResult) {
   uint32_t i, last;
   last = mStartedReadingCurrent ? mCurrentStream + 1 : mCurrentStream;
   for (i = 0; i < last; ++i) {
-    if (NS_WARN_IF(!mStreams[i].mSeekableStream)) {
+    if (NS_WARN_IF(!mStreams[i].mTellableStream)) {
       return NS_ERROR_NO_INTERFACE;
     }
 
     int64_t pos;
-    rv = TellMaybeSeek(mStreams[i].mSeekableStream, &pos);
+    rv = TellMaybeSeek(mStreams[i].mTellableStream, mStreams[i].mSeekableStream,
+                       &pos);
     if (NS_WARN_IF(NS_FAILED(rv))) {
       return rv;
     }
@@ -910,23 +991,72 @@ nsresult nsMultiplexInputStreamConstructor(nsISupports* aOuter, REFNSIID aIID,
 }
 
 void nsMultiplexInputStream::Serialize(InputStreamParams& aParams,
-                                       FileDescriptorArray& aFileDescriptors) {
+                                       FileDescriptorArray& aFileDescriptors,
+                                       bool aDelayedStart, uint32_t aMaxSize,
+                                       uint32_t* aSizeUsed,
+                                       mozilla::dom::ContentChild* aManager) {
+  SerializeInternal(aParams, aFileDescriptors, aDelayedStart, aMaxSize,
+                    aSizeUsed, aManager);
+}
+
+void nsMultiplexInputStream::Serialize(InputStreamParams& aParams,
+                                       FileDescriptorArray& aFileDescriptors,
+                                       bool aDelayedStart, uint32_t aMaxSize,
+                                       uint32_t* aSizeUsed,
+                                       PBackgroundChild* aManager) {
+  SerializeInternal(aParams, aFileDescriptors, aDelayedStart, aMaxSize,
+                    aSizeUsed, aManager);
+}
+
+void nsMultiplexInputStream::Serialize(InputStreamParams& aParams,
+                                       FileDescriptorArray& aFileDescriptors,
+                                       bool aDelayedStart, uint32_t aMaxSize,
+                                       uint32_t* aSizeUsed,
+                                       mozilla::dom::ContentParent* aManager) {
+  SerializeInternal(aParams, aFileDescriptors, aDelayedStart, aMaxSize,
+                    aSizeUsed, aManager);
+}
+
+void nsMultiplexInputStream::Serialize(InputStreamParams& aParams,
+                                       FileDescriptorArray& aFileDescriptors,
+                                       bool aDelayedStart, uint32_t aMaxSize,
+                                       uint32_t* aSizeUsed,
+                                       PBackgroundParent* aManager) {
+  SerializeInternal(aParams, aFileDescriptors, aDelayedStart, aMaxSize,
+                    aSizeUsed, aManager);
+}
+
+template <typename M>
+void nsMultiplexInputStream::SerializeInternal(
+    InputStreamParams& aParams, FileDescriptorArray& aFileDescriptors,
+    bool aDelayedStart, uint32_t aMaxSize, uint32_t* aSizeUsed, M* aManager) {
   MutexAutoLock lock(mLock);
 
   MultiplexInputStreamParams params;
 
-  uint32_t streamCount = mStreams.Length();
+  CheckedUint32 totalSizeUsed = 0;
+  CheckedUint32 maxSize = aMaxSize;
 
+  uint32_t streamCount = mStreams.Length();
   if (streamCount) {
     InfallibleTArray<InputStreamParams>& streams = params.streams();
 
     streams.SetCapacity(streamCount);
     for (uint32_t index = 0; index < streamCount; index++) {
+      uint32_t sizeUsed = 0;
       InputStreamParams childStreamParams;
       InputStreamHelper::SerializeInputStream(
-          mStreams[index].mStream, childStreamParams, aFileDescriptors);
-
+          mStreams[index].mStream, childStreamParams, aFileDescriptors,
+          aDelayedStart, maxSize.value(), &sizeUsed, aManager);
       streams.AppendElement(childStreamParams);
+
+      MOZ_ASSERT(maxSize.value() >= sizeUsed);
+
+      maxSize -= sizeUsed;
+      MOZ_DIAGNOSTIC_ASSERT(maxSize.isValid());
+
+      totalSizeUsed += sizeUsed;
+      MOZ_DIAGNOSTIC_ASSERT(totalSizeUsed.isValid());
     }
   }
 
@@ -935,6 +1065,9 @@ void nsMultiplexInputStream::Serialize(InputStreamParams& aParams,
   params.startedReadingCurrent() = mStartedReadingCurrent;
 
   aParams = params;
+
+  MOZ_ASSERT(aSizeUsed);
+  *aSizeUsed = totalSizeUsed.value();
 }
 
 bool nsMultiplexInputStream::Deserialize(
@@ -972,33 +1105,11 @@ bool nsMultiplexInputStream::Deserialize(
   return true;
 }
 
-Maybe<uint64_t> nsMultiplexInputStream::ExpectedSerializedLength() {
-  MutexAutoLock lock(mLock);
-
-  bool lengthValueExists = false;
-  uint64_t expectedLength = 0;
-  uint32_t streamCount = mStreams.Length();
-  for (uint32_t index = 0; index < streamCount; index++) {
-    nsCOMPtr<nsIIPCSerializableInputStream> stream =
-        do_QueryInterface(mStreams[index].mStream);
-    if (!stream) {
-      continue;
-    }
-    Maybe<uint64_t> length = stream->ExpectedSerializedLength();
-    if (length.isNothing()) {
-      continue;
-    }
-    lengthValueExists = true;
-    expectedLength += length.value();
-  }
-  return lengthValueExists ? Some(expectedLength) : Nothing();
-}
-
 NS_IMETHODIMP
 nsMultiplexInputStream::GetCloneable(bool* aCloneable) {
   MutexAutoLock lock(mLock);
   // XXXnsm Cloning a multiplex stream which has started reading is not
-  // permitted  right now.
+  // permitted right now.
   if (mCurrentStream > 0 || mStartedReadingCurrent) {
     *aCloneable = false;
     return NS_OK;
@@ -1023,7 +1134,7 @@ nsMultiplexInputStream::Clone(nsIInputStream** aClone) {
   MutexAutoLock lock(mLock);
 
   // XXXnsm Cloning a multiplex stream which has started reading is not
-  // permitted  right now.
+  // permitted right now.
   if (mCurrentStream > 0 || mStartedReadingCurrent) {
     return NS_ERROR_FAILURE;
   }
@@ -1055,6 +1166,301 @@ nsMultiplexInputStream::Clone(nsIInputStream** aClone) {
   return NS_OK;
 }
 
+NS_IMETHODIMP
+nsMultiplexInputStream::Length(int64_t* aLength) {
+  MutexAutoLock lock(mLock);
+
+  if (mCurrentStream > 0 || mStartedReadingCurrent) {
+    return NS_ERROR_NOT_AVAILABLE;
+  }
+
+  CheckedInt64 length = 0;
+  nsresult retval = NS_OK;
+
+  for (uint32_t i = 0, len = mStreams.Length(); i < len; ++i) {
+    nsCOMPtr<nsIInputStreamLength> substream =
+        do_QueryInterface(mStreams[i].mStream);
+    if (!substream) {
+      // Let's use available as fallback.
+      uint64_t streamAvail = 0;
+      nsresult rv = AvailableMaybeSeek(mStreams[i], &streamAvail);
+      if (rv == NS_BASE_STREAM_CLOSED) {
+        continue;
+      }
+
+      if (NS_WARN_IF(NS_FAILED(rv))) {
+        mStatus = rv;
+        return mStatus;
+      }
+
+      length += streamAvail;
+      if (!length.isValid()) {
+        return NS_ERROR_OUT_OF_MEMORY;
+      }
+
+      continue;
+    }
+
+    int64_t size = 0;
+    nsresult rv = substream->Length(&size);
+    if (rv == NS_BASE_STREAM_CLOSED) {
+      continue;
+    }
+
+    if (rv == NS_ERROR_NOT_AVAILABLE) {
+      return rv;
+    }
+
+    // If one stream blocks, we all block.
+    if (rv != NS_BASE_STREAM_WOULD_BLOCK && NS_WARN_IF(NS_FAILED(rv))) {
+      return rv;
+    }
+
+    // We want to return WOULD_BLOCK if there is 1 stream that blocks. But want
+    // to see if there are other streams with length = -1.
+    if (rv == NS_BASE_STREAM_WOULD_BLOCK) {
+      retval = NS_BASE_STREAM_WOULD_BLOCK;
+      continue;
+    }
+
+    // If one of the stream doesn't know the size, we all don't know the size.
+    if (size == -1) {
+      *aLength = -1;
+      return NS_OK;
+    }
+
+    length += size;
+    if (!length.isValid()) {
+      return NS_ERROR_OUT_OF_MEMORY;
+    }
+  }
+
+  *aLength = length.value();
+  return retval;
+}
+
+class nsMultiplexInputStream::AsyncWaitLengthHelper final
+    : public nsIInputStreamLengthCallback
+
+{
+ public:
+  NS_DECL_ISUPPORTS
+
+  AsyncWaitLengthHelper()
+      : mStreamNotified(false), mLength(0), mNegativeSize(false) {}
+
+  bool AddStream(nsIAsyncInputStreamLength* aStream) {
+    return mPendingStreams.AppendElement(aStream, fallible);
+  }
+
+  bool AddSize(int64_t aSize) {
+    MOZ_ASSERT(!mNegativeSize);
+
+    mLength += aSize;
+    return mLength.isValid();
+  }
+
+  void NegativeSize() {
+    MOZ_ASSERT(!mNegativeSize);
+    mNegativeSize = true;
+  }
+
+  nsresult Proceed(nsMultiplexInputStream* aParentStream,
+                   nsIEventTarget* aEventTarget,
+                   const MutexAutoLock& aProofOfLock) {
+    MOZ_ASSERT(!mStream);
+
+    // If we don't need to wait, let's inform the callback immediately.
+    if (mPendingStreams.IsEmpty() || mNegativeSize) {
+      RefPtr<nsMultiplexInputStream> parentStream = aParentStream;
+      int64_t length = -1;
+      if (!mNegativeSize && mLength.isValid()) {
+        length = mLength.value();
+      }
+      nsCOMPtr<nsIRunnable> r = NS_NewRunnableFunction(
+          "AsyncWaitLengthHelper", [parentStream, length]() {
+            MutexAutoLock lock(parentStream->GetLock());
+            parentStream->AsyncWaitCompleted(length, lock);
+          });
+      return aEventTarget->Dispatch(r.forget(), NS_DISPATCH_NORMAL);
+    }
+
+    // Let's store the callback and the parent stream until we have
+    // notifications from the async length streams.
+
+    mStream = aParentStream;
+
+    // Let's activate all the pending streams.
+    for (nsIAsyncInputStreamLength* stream : mPendingStreams) {
+      nsresult rv = stream->AsyncLengthWait(this, aEventTarget);
+      if (rv == NS_BASE_STREAM_CLOSED) {
+        continue;
+      }
+
+      if (NS_WARN_IF(NS_FAILED(rv))) {
+        return rv;
+      }
+    }
+
+    return NS_OK;
+  }
+
+  NS_IMETHOD
+  OnInputStreamLengthReady(nsIAsyncInputStreamLength* aStream,
+                           int64_t aLength) override {
+    MutexAutoLock lock(mStream->GetLock());
+
+    MOZ_ASSERT(mPendingStreams.Contains(aStream));
+    mPendingStreams.RemoveElement(aStream);
+
+    // Already notified.
+    if (mStreamNotified) {
+      return NS_OK;
+    }
+
+    if (aLength == -1) {
+      mNegativeSize = true;
+    } else {
+      mLength += aLength;
+      if (!mLength.isValid()) {
+        mNegativeSize = true;
+      }
+    }
+
+    // We need to wait.
+    if (!mNegativeSize && !mPendingStreams.IsEmpty()) {
+      return NS_OK;
+    }
+
+    // Let's notify the parent stream.
+    mStreamNotified = true;
+    mStream->AsyncWaitCompleted(mNegativeSize ? -1 : mLength.value(), lock);
+    return NS_OK;
+  }
+
+ private:
+  ~AsyncWaitLengthHelper() = default;
+
+  RefPtr<nsMultiplexInputStream> mStream;
+  bool mStreamNotified;
+
+  CheckedInt64 mLength;
+  bool mNegativeSize;
+
+  nsTArray<nsCOMPtr<nsIAsyncInputStreamLength>> mPendingStreams;
+};
+
+NS_IMPL_ISUPPORTS(nsMultiplexInputStream::AsyncWaitLengthHelper,
+                  nsIInputStreamLengthCallback)
+
+NS_IMETHODIMP
+nsMultiplexInputStream::AsyncLengthWait(nsIInputStreamLengthCallback* aCallback,
+                                        nsIEventTarget* aEventTarget) {
+  if (NS_WARN_IF(!aEventTarget)) {
+    return NS_ERROR_NULL_POINTER;
+  }
+
+  MutexAutoLock lock(mLock);
+
+  if (mCurrentStream > 0 || mStartedReadingCurrent) {
+    return NS_ERROR_NOT_AVAILABLE;
+  }
+
+  if (!aCallback) {
+    mAsyncWaitLengthCallback = nullptr;
+    return NS_OK;
+  }
+
+  // We have a pending operation! Let's use this instead of creating a new one.
+  if (mAsyncWaitLengthHelper) {
+    mAsyncWaitLengthCallback = aCallback;
+    return NS_OK;
+  }
+
+  RefPtr<AsyncWaitLengthHelper> helper = new AsyncWaitLengthHelper();
+
+  for (uint32_t i = 0, len = mStreams.Length(); i < len; ++i) {
+    nsCOMPtr<nsIAsyncInputStreamLength> asyncStream =
+        do_QueryInterface(mStreams[i].mStream);
+    if (asyncStream) {
+      if (NS_WARN_IF(!helper->AddStream(asyncStream))) {
+        return NS_ERROR_OUT_OF_MEMORY;
+      }
+      continue;
+    }
+
+    nsCOMPtr<nsIInputStreamLength> stream =
+        do_QueryInterface(mStreams[i].mStream);
+    if (!stream) {
+      // Let's use available as fallback.
+      uint64_t streamAvail = 0;
+      nsresult rv = AvailableMaybeSeek(mStreams[i], &streamAvail);
+      if (rv == NS_BASE_STREAM_CLOSED) {
+        continue;
+      }
+
+      if (NS_WARN_IF(NS_FAILED(rv))) {
+        mStatus = rv;
+        return mStatus;
+      }
+
+      if (NS_WARN_IF(!helper->AddSize(streamAvail))) {
+        return NS_ERROR_OUT_OF_MEMORY;
+      }
+
+      continue;
+    }
+
+    int64_t size = 0;
+    nsresult rv = stream->Length(&size);
+    if (rv == NS_BASE_STREAM_CLOSED) {
+      continue;
+    }
+
+    MOZ_ASSERT(rv != NS_BASE_STREAM_WOULD_BLOCK,
+               "A nsILengthInutStream returns NS_BASE_STREAM_WOULD_BLOCK but "
+               "it doesn't implement nsIAsyncInputStreamLength.");
+
+    if (NS_WARN_IF(NS_FAILED(rv))) {
+      return rv;
+    }
+
+    if (size == -1) {
+      helper->NegativeSize();
+      break;
+    }
+
+    if (NS_WARN_IF(!helper->AddSize(size))) {
+      return NS_ERROR_OUT_OF_MEMORY;
+    }
+  }
+
+  nsresult rv = helper->Proceed(this, aEventTarget, lock);
+  if (NS_WARN_IF(NS_FAILED(rv))) {
+    return rv;
+  }
+
+  mAsyncWaitLengthHelper = helper;
+  mAsyncWaitLengthCallback = aCallback;
+  return NS_OK;
+}
+
+void nsMultiplexInputStream::AsyncWaitCompleted(
+    int64_t aLength, const MutexAutoLock& aProofOfLock) {
+  nsCOMPtr<nsIInputStreamLengthCallback> callback;
+  callback.swap(mAsyncWaitLengthCallback);
+
+  mAsyncWaitLengthHelper = nullptr;
+
+  // Already canceled.
+  if (!callback) {
+    return;
+  }
+
+  MutexAutoUnlock unlock(mLock);
+  callback->OnInputStreamLengthReady(this, aLength);
+}
+
 #define MAYBE_UPDATE_VALUE_REAL(x, y)                         \
   if (y) {                                                    \
     if (aCount == 1) {                                        \
@@ -1078,9 +1484,12 @@ void nsMultiplexInputStream::UpdateQIMap(StreamData& aStream, int32_t aCount) {
   MOZ_ASSERT(aCount == -1 || aCount == 1);
 
   MAYBE_UPDATE_VALUE_REAL(mSeekableStreams, aStream.mSeekableStream)
+  MAYBE_UPDATE_VALUE_REAL(mTellableStreams, aStream.mTellableStream)
   MAYBE_UPDATE_VALUE(mIPCSerializableStreams, nsIIPCSerializableInputStream)
   MAYBE_UPDATE_VALUE(mCloneableStreams, nsICloneableInputStream)
   MAYBE_UPDATE_VALUE_REAL(mAsyncInputStreams, aStream.mAsyncStream)
+  MAYBE_UPDATE_VALUE(mInputStreamLengths, nsIInputStreamLength)
+  MAYBE_UPDATE_VALUE(mAsyncInputStreamLengths, nsIAsyncInputStreamLength)
 }
 
 #undef MAYBE_UPDATE_VALUE
@@ -1089,6 +1498,10 @@ bool nsMultiplexInputStream::IsSeekable() const {
   return mStreams.Length() == mSeekableStreams;
 }
 
+bool nsMultiplexInputStream::IsTellable() const {
+  return mStreams.Length() == mTellableStreams;
+}
+
 bool nsMultiplexInputStream::IsIPCSerializable() const {
   return mStreams.Length() == mIPCSerializableStreams;
 }
@@ -1102,3 +1515,11 @@ bool nsMultiplexInputStream::IsAsyncInputStream() const {
   // substream implements that interface.
   return !!mAsyncInputStreams;
 }
+
+bool nsMultiplexInputStream::IsInputStreamLength() const {
+  return !!mInputStreamLengths;
+}
+
+bool nsMultiplexInputStream::IsAsyncInputStreamLength() const {
+  return !!mAsyncInputStreamLengths;
+}
